from math import sqrt

G = 6.673e-11  # gravitational constant


class Body:
    # cartesian positions
    rx: float
    ry: float

    # velocity components
    vx: float
    vy: float

    # force components
    fx: float = 0.0
    fy: float = 0.0

    mass: float

    def __init__(self, rx: float, ry: float,
                 vx: float, vy: float,
                 mass: float):
        self.rx = rx
        self.ry = ry
        self.vx = vx
        self.vy = vy
        self.mass = mass

    def __str__(self):
        return "{}, {}, {}, {}, {}".format(
            self.rx, self.ry, self.vx, self.vy, self.mass
        )

    def update(self, dt: float):
        """
        Update the velocity and position using a time step dt
        """
        self.vx += dt * self.fx / self.mass
        self.vy += dt * self.fy / self.mass
        self.rx += dt * self.vx
        self.ry += dt * self.vy

    def distance_to(self, other_body: 'Body') -> float:
        """'
        Return the distance between two bodies
        """
        dx = self.rx - other_body.rx
        dy = self.ry - other_body.ry
        return sqrt(dx*dx + dy*dy)

    def reset_force(self):
        """
        Reset the force to 0 for the next iteration
        """
        self.fx = 0.0
        self.fy = 0.0

    def accelerate(self, other_body: 'Body'):
        """
        Compute the net force acting between this body and
        other_body and add to the net force acting on this
        body
        """

        # These two checks are not in the original source
        if self == other_body:
            return

        if other_body is None:
            return

        # softening parameter (just to avoid infinities)
        eps = 3E4
        dx = other_body.rx - self.rx
        dy = other_body.ry - self.ry
        dist = sqrt(dx*dx + dy*dy)
        F = (G * self.mass * other_body.mass) / (dist*dist + eps*eps)
        self.fx = self.fx + F * dx / dist
        self.fy = self.fy + F * dy / dist

    def is_in(self, quad: 'Quadrant'):
        return quad.contains(self.rx, self.ry)


from .simulation import *

if __name__ == '__main__':
    print('*' * 50)
    
    system = None
    if simulation_type == 'BarnesHut':
        system = RenderableBarnesHutSystem()
    elif simulation_type == 'BruteForce':
        system = RenderableBruteForceSystem()

    system.start_the_bodies(int(bodies))
    renderer = SystemRenderer(system, frames=int(frames), trail_size=int(trail_size), performance_test=performance_test)

    renderer.run()




from .body import Body
from .quadrant import Quadrant


class BHTree:
    body: Body = None
    nw: 'BHTree' = None
    ne: 'BHTree' = None
    sw: 'BHTree' = None
    se: 'BHTree' = None

    def __init__(self, quad: Quadrant):
        self.quadrant = quad

    def is_leaf(self) -> bool:
        """
        If all nodes of the BHTree are null, then the quadrant represents a single body and
        it is "external" or a leaf of the tree
        """
        return self.nw is None and self.ne is None and self.sw is None and self.se is None

    def insert(self, b: Body):
        """
        We have to populate the tree with bodies. We start at the current tree and
        recursively travel through the branches
        """
        # If there's not a b there already, put the b there.
        if not self.body:
            self.body = b

        # If there's already a body there, but it's not an external node
        # combine the two bodies and figure out which quadrant of the
        # tree it should be located in. Then recursively update the nodes below
        elif not self.is_leaf():
            # In this case, this means "add the acting forces together
            b.accelerate(self.body)

            northwest = self.quadrant.NW()
            if b.is_in(northwest):
                if self.nw is None:
                    self.nw = BHTree(northwest)
                self.nw.insert(b)
                return

            northeast = self.quadrant.NE()
            if b.is_in(northeast):
                if self.ne is None:
                    self.ne = BHTree(northeast)
                self.ne.insert(b)
                return

            southeast = self.quadrant.SE()
            if b.is_in(southeast):
                if self.se is None:
                    self.se = BHTree(northeast)
                self.se.insert(b)
                return

            southwest = self.quadrant.SW()
            if b.is_in(southwest):
                if self.sw is None:
                    self.sw = BHTree(southwest)
                self.sw.insert(b)
                return

        # If the node is a leaf and contains another body, create BHTrees
        # where the bodies should go, update the node, and end
        # (do not do anything recursively)
        else:
            c = self.body
            northwest = self.quadrant.NW()
            if c.is_in(northwest):
                if self.nw is None:
                    self.nw = BHTree(northwest)
                self.insert(c)
            northeast = self.quadrant.NE()
            if c.is_in(northeast):
                if self.ne is None:
                    self.ne = BHTree(northeast)
                self.insert(c)
            southwest = self.quadrant.SW()
            if c.is_in(southwest):
                if self.sw is None:
                    self.sw = BHTree(southwest)
                self.insert(c)
            southeast = self.quadrant.SE()
            if c.is_in(southeast):
                if self.se is None:
                    self.se = BHTree(southeast)
                self.insert(c)

    def update_force(self, b: Body):
        """
        Start at the main node of the tree. Then, recursively go each branch
        Until either we reach an leaf node or we reach a node that is sufficiently
        far away that the external nodes would not matter much.
        """

        if self.is_leaf():
            b.accelerate(self.body)
        elif self.body != b and self.quadrant.diameter / self.body.distance_to(b) < 2:
            b.accelerate(self.body)
        else:
            for quadrant in [self.nw, self.ne, self.sw, self.se]:
                if quadrant is not None:
                    quadrant.update_force(b)

    def __str__(self):
        if self.ne is None or self.nw is None or self.se is None or self.sw is None:
            return f'* {self.body}\n{self.nw}{self.ne}{self.sw}{self.se}'
        else:
            return f'  {self.body}\n'


import argparse
from .rendering import *
from os import path, mkdir

if not path.exists("./renders"):
    mkdir("./renders")

def main():
    parser = argparse.ArgumentParser(description='Simulate N-Body Problem')
    parser.add_argument("simulation_type", help="BruteForce or BarnesHut")
    parser.add_argument("bodies", help="The number of bodies to simulate")
    parser.add_argument("frames", help="The number of frames to simulate for")
    parser.add_argument("trail_size", help="Display a trail of diameter x while rendering")
    parser.add_argument("--performance_test", help="don't render anything, just calculate", action="store_true")
    args = parser.parse_args()

    system = None
    if args.simulation_type == 'BarnesHut':
        system = RenderableBarnesHutSystem()
    elif args.simulation_type == 'BruteForce':
        system = RenderableBruteForceSystem()
    else:
        exit(1)

    system.start_the_bodies(int(args.bodies))
    renderer = SystemRenderer(system, frames=int(args.frames), trail_size=int(args.trail_size), performance_test=args.performance_test)

    renderer.run()
    


import setuptools
import CrudeBHT

def requirements():
    with open('requirements.txt') as f:
        install_requires = [line.strip() for line in f]
    return install_requires

with open("readme.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "CrudeBHT",
    version = "0.0.7",
    author = "Naimish Mani B",
    author_email = "naimish240@gmail.com",
    description = "A simple implementation of the Barnes-Hut algorithm for N-body simulations in python 3.7 ",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    entry_points = {'console_scripts' : ['crudebht = CrudeBHT.__main__:main']},
    packages=setuptools.find_packages(),
    install_require = requirements(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


from math import sqrt, exp, atan, pi, cos, copysign, sin, log
from random import random

from .bhtree import BHTree
from .body import Body
from .quadrant import Quadrant


class System:
    """A container for orbital dynamics simulations"""
    bodies: [Body] = []
    solar_mass = 1.98892e30
    radius = 1e18  # the radius of the system
    time_warp_factor = 2e10

    def parent_body_color(self):
        return '#FDB813'

    @property
    def n(self):
        return len(self.bodies)

    def __init__(self):
        pass

    def calculate(self, elapsed_time):
        self.accelerate(self.n, elapsed_time)

    def circular_velocity(self, rx: float, ry: float) -> float:
        """
        the bodies are initialized in circular orbits around the central mass.
        This is just some physics to do that
        """
        r2 = sqrt(rx*rx + ry*ry)
        numerator = 6.67e-11 * 1e6 * self.solar_mass
        return sqrt(numerator/r2)

    def start_the_bodies(self, n: int):
        """
        Initialize N bodies with random positions and circular velocities
        """
        # Put something heavy into the center
        self.bodies.append(Body(0, 0, 0, 0, 1e6 * self.solar_mass))

        # Put some bodies around it
        for i in range(n):
            px = (self.radius + 1e8) * exp(-1.8)  * (0.5-random())
            py = (self.radius + 1e8) * exp(-1.8) * (0.5-random())
            magv = self.circular_velocity(px, py) * (0.7 + random() * 0.5)

            absangle = atan(abs(py / px))
            thetav = pi / 2 - absangle

            vx = -1 * copysign(1, py) * cos(thetav) * magv
            vy = copysign(1, px) * sin(thetav) * magv

            # Most objects should orbit in one direction, but there may be exceptions
            if random() <= 0.6:
                vx = -vx
                vy = -vy

            mass = random() * self.solar_mass * 30 + 1e20
            self.bodies.append(Body(px, py, vx, vy, mass))

    def accelerate(self, n, elapsed_time):
        """Override this method in your subclass"""
        raise NotImplementedError

    def exp(self, lmbda):
        """
        A function to return an exponential distribution for position
        """
        return -log(1 - random()) / lmbda


class BruteForceSystem(System):
    """
    Naive implementation of orbital dynamics
    """
    def accelerate(self, n, elapsed_time):
        for body in self.bodies:
            body.reset_force()

            # This'll get us n^2 complexity
            for other_body in self.bodies:
                if other_body != body:
                    body.accelerate(other_body)

        # Update the timestamps
        for body in self.bodies:
            body.update(elapsed_time)

class BarnesHutSystem(System):
    """
    System that uses a Barnes-Hut tree to calculte orbital dynamics
    """
    quad = Quadrant(0, 0, 2 * 1e18)

    def accelerate(self, n, elapsed_time):
        the_tree = BHTree(self.quad)

        # If the body is inside the current quadrant, add it to the tree
        for body in self.bodies:
            if body.is_in(self.quad):
                the_tree.insert(body)

        # Now, use our methods in BHTree to update the forces,
        # traveling recursively through the tree
        for body in self.bodies:
            body.reset_force()
            if body.is_in(self.quad):
                the_tree.update_force(body)
                # Calculate the new positions
                body.update(elapsed_time)


class Quadrant:
    """
    A Quadrant describes a square region in space. In our two dimensional simulation, a quadrant
    has x and y coordinates for its center, as well as a diameter property that describes its width
    and height
    """
    def __init__(self, xmid: float, ymid: float, diameter: float):
        self.xmid = xmid
        self.ymid = ymid
        self.diameter = diameter

    def contains(self, xmid, ymid):
        """
        Check if this Quadrant contains a point
        """
        if (xmid <= self.xmid + self.diameter/2.0) and \
           (xmid >= self.xmid - self.diameter/2.0) and \
           (ymid <= self.ymid + self.diameter/2.0) and \
           (ymid >= self.ymid - self.diameter/2.0):
            return True
        return False

    def NW(self) -> 'Quadrant':
        return Quadrant(self.xmid - self.diameter / 4.0, self.ymid + self.diameter / 4.0, self.diameter / 2.0)

    def NE(self) -> 'Quadrant':
        return Quadrant(self.xmid + self.diameter / 4.0, self.ymid + self.diameter / 4.0, self.diameter / 2.0)

    def SW(self) -> 'Quadrant':
        return Quadrant(self.xmid - self.diameter / 4.0, self.ymid - self.diameter / 4.0, self.diameter / 2.0)

    def SE(self) -> 'Quadrant':
        return Quadrant(self.xmid + self.diameter / 4.0, self.ymid - self.diameter / 4.0, self.diameter / 2.0)


from builtins import int
from typing import Tuple, List
from tqdm import tqdm

import math

import matplotlib.pyplot as plt
import matplotlib.animation as animation
Writer = animation.FFMpegWriter(fps=30, codec='libx264')  # To get the animation working

import time

from .body import Body
from .system import BruteForceSystem, BarnesHutSystem

CURSOR_UP_ONE = '\x1b[1A'
ERASE_LINE = '\x1b[2K'


class RenderableSystem:
    def name(self) -> str:
        raise NotImplementedError

    def get_solar_mass(self) -> float:
        raise NotImplementedError

    def get_x_limits(self) -> Tuple[int, int]:
        raise NotImplementedError

    def get_y_limits(self) -> Tuple[int, int]:
        raise NotImplementedError

    def get_bodies(self) -> List[Body]:
        raise NotImplementedError

    def step(self, elapsed_time):
        raise NotImplementedError


class SystemRenderer:
    min_marker_size = 0.1
    max_marker_size = 10
    progress_bar = None

    _background = None
    _body_dots = []
    _body_trailing_lines = []

    __last_step_time = None

    @property
    def time_warp_factor(self):
        return self.system.time_warp_factor

    @property
    def _last_step_time(self):
        if self.__last_step_time is None:
            self.__last_step_time = time.process_time()
        return self.__last_step_time

    def __init__(self, universe: RenderableSystem, frames: int, trail_size: int = 0,
                 performance_test: bool=False):
        self.system = universe
        self.trail_size = trail_size
        self.frames = frames
        self.performance_test = performance_test

        if self.performance_test:
            return

        self._fig, self._ax = plt.subplots(1, 1, figsize=(16, 12))
        self._ax.set_aspect('equal')
        self._ax.set_xlim(*self.system.get_x_limits())
        self._ax.set_ylim(*self.system.get_y_limits())

        # Comment to show axies
        plt.axis('off')

        # Setup Dots and Lines
        for body in self.system.get_bodies():
            marker_size = (math.floor(body.mass / self.system.get_solar_mass() / 2) + 1) 
            marker_size = min(marker_size, self.max_marker_size)
            marker_size = max(marker_size, self.min_marker_size)
            dot = plt.plot(body.rx, body.ry, 'o', markersize=marker_size)

            if body == self.system.get_bodies()[0]:
                # Special Case for the sun
                dot[0].set_color(self.system.parent_body_color())

            self._body_dots.append(dot)

            trailing_lines = []
            for _ in range(self.trail_size):
                trailing_line = plt.plot([0, 0], [0, 0], dot[0].get_color())
                trailing_lines.append(trailing_line)
            self._body_trailing_lines.append(trailing_lines)

    def run(self):
        self.progress_bar = tqdm(total=self.frames, unit=' frames')
        if self.performance_test:
            for frame in range(self.frames):
                self.progress_bar.update()
                self.system.step(self.time_warp_factor)
            self.progress_bar.update()
        else:
            file_name = f'renders/render-{self.system.name()}-{len(self.system.get_bodies())}b-{self.frames}f.mp4'
            anim = animation.FuncAnimation(self._fig, self.step, self.frames, interval=30, blit=True)
            anim.save(file_name)
        self.progress_bar.close()

    def step(self, frame):
        self.progress_bar.update()
        self.system.step(self.time_warp_factor)

        # Update The Dots
        dots_to_update = []
        for i in range(len(self.system.get_bodies())):
            dot = self._body_dots[i][0]
            body = self.system.get_bodies()[i]
            dots_to_update.append(dot)

            old_xdata = dot.get_xdata()
            old_ydata = dot.get_ydata()
            xdata = [body.rx]
            ydata = [body.ry]

            # Update the Dot
            dot.set_xdata(xdata)
            dot.set_ydata(ydata)

            # Update the trailing line
            if self.trail_size < 1:
                continue
            trailing_line = self._body_trailing_lines[i].pop()
            trailing_line[0].set_xdata([old_xdata, xdata])
            trailing_line[0].set_ydata([old_ydata, ydata])
            self._body_trailing_lines[i].insert(0, trailing_line)
        return dots_to_update


class RenderableBruteForceSystem(BruteForceSystem, RenderableSystem):
    def name(self):
        return 'BruteForce'

    def get_solar_mass(self):
        return self.solar_mass

    def get_x_limits(self):
        return -self.radius/6, self.radius/6

    def get_y_limits(self):
        return -self.radius/8, self.radius/8

    def get_bodies(self):
        return self.bodies

    def step(self, elapsed_time):
        self.calculate(elapsed_time)

class RenderableBarnesHutSystem(BarnesHutSystem, RenderableSystem):
    def name(self):
        return 'BarnesHut'

    def get_solar_mass(self):
        return self.solar_mass

    def get_x_limits(self):
        return -self.radius/6, self.radius/6

    def get_y_limits(self):
        return -self.radius/8, self.radius/8

    def get_bodies(self):
        return self.bodies

    def step(self, elapsed_time):
        self.calculate(elapsed_time)